# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>
# pylint: disable=protected-access, unused-argument, missing-function-docstring

""" Unit tests for the hebg.behavior module. """

from copy import deepcopy
from networkx.algorithms.shortest_paths.unweighted import predecessor

import pytest
import pytest_check as check
from pytest_mock import MockerFixture

from hebg.heb_graph import HEBGraph, DiGraph, Behavior


class TestHEBGraph:

    """HEBGraph"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""
        self.behavior = Behavior("base_behavior_name")
        self.heb_graph = HEBGraph(self.behavior)

    def test_init(self):
        """should instanciate correctly."""
        graph = self.heb_graph
        check.equal(graph.behavior, self.behavior)
        check.equal(graph.all_behaviors, {})
        check.equal(graph.any_mode, "first")
        check.is_true(isinstance(graph, DiGraph))

    def test_add_node(self, mocker: MockerFixture):
        """should add Node to the graph correctly."""
        mocker.patch("hebg.heb_graph.DiGraph.add_node")

        class DummyNode:
            """DummyNode"""

            name = "node_name"
            type = "node_type"
            image = "node_image"

            def __str__(self) -> str:
                return self.name

        node = DummyNode()
        expected_kwargs = {"type": "node_type", "image": "node_image", "color": None}
        expected_args = (node,)

        self.heb_graph.add_node(node)
        args, kwargs = DiGraph.add_node.call_args
        check.equal(kwargs, expected_kwargs)
        check.equal(args, expected_args)

    def test_add_edge_only(self, mocker: MockerFixture):
        """should add edges to the graph correctly if nodes already exists."""
        mocker.patch("hebg.heb_graph.DiGraph.add_edge")

        class DummyNode:
            """DummyNode"""

            type = "node_type"
            image = "node_image"

            def __init__(self, i) -> None:
                self.name = f"node_name_{i}"

            def __str__(self) -> str:
                return self.name

        node_0, node_1 = DummyNode(0), DummyNode(1)
        self.heb_graph.add_node(node_0)
        self.heb_graph.add_node(node_1)

        mocker.patch("hebg.heb_graph.DiGraph.add_node")
        expected_kwargs = {"index": 42, "color": "black"}
        expected_args = (node_0, node_1)

        self.heb_graph.add_edge(node_0, node_1, index=42)
        args, kwargs = DiGraph.add_edge.call_args
        check.equal(kwargs, expected_kwargs)
        check.equal(args, expected_args)
        check.is_false(DiGraph.add_node.called)

    def test_add_edge_and_nodes(self, mocker: MockerFixture):
        """should add edges and nodes to the graph correctly if nodes are not in the graph yet."""
        mocker.patch("hebg.heb_graph.DiGraph.add_edge")
        mocker.patch("hebg.heb_graph.DiGraph.add_node")

        class DummyNode:
            """DummyNode"""

            type = "node_type"
            image = "node_image"

            def __init__(self, i) -> None:
                self.name = f"node_name_{i}"

            def __str__(self) -> str:
                return self.name

        node_0, node_1 = DummyNode(0), DummyNode(1)

        expected_nodes_args = ((node_0,), (node_1,))
        expected_edge_kwargs = {"index": 42, "color": "black"}
        expected_edge_args = (node_0, node_1)

        self.heb_graph.add_edge(node_0, node_1, index=42)
        args, kwargs = DiGraph.add_edge.call_args
        check.equal(kwargs, expected_edge_kwargs)
        check.equal(args, expected_edge_args)

        for i, (args, _) in enumerate(DiGraph.add_node.call_args_list):
            check.equal(args, expected_nodes_args[i])

    def test_call(self, mocker: MockerFixture):
        """should return roots action on call."""
        roots = "roots"
        observation = "obs"
        mocker.patch("hebg.heb_graph.HEBGraph._get_any_action")
        mocker.patch("hebg.heb_graph.HEBGraph.roots", roots)
        self.heb_graph(observation)
        args, _ = HEBGraph._get_any_action.call_args
        check.equal(args[0], roots)
        check.equal(args[1], observation)
        check.equal(args[2], [self.behavior.name])

    def test_roots(self, mocker: MockerFixture):
        """should have roots as property."""

        nodes = ["A", "B", "C", "AA", "AB"]
        predecessors = {"A": [], "B": [], "C": [], "AA": ["A"], "AB": ["A", "B"]}

        mocker.patch("hebg.heb_graph.HEBGraph.nodes", lambda self: nodes)
        mocker.patch(
            "hebg.heb_graph.HEBGraph.predecessors",
            lambda self, node: predecessors[node],
        )

        check.equal(self.heb_graph.roots, ["A", "B", "C"])


class TestHEBGraphGetAnyAction:

    """HEBGraph._get_any_action"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""
        self.behavior = Behavior("behavior_name")
        self.heb_graph = HEBGraph(self.behavior)

    def test_none_in_actions(self, mocker: MockerFixture):
        """should return None if any node returns None."""

        actions = [0, "Impossible", 2, None, 3]

        _actions = deepcopy(actions)

        def mocked_get_action(*args):
            return _actions.pop(0)

        mocker.patch("hebg.heb_graph.HEBGraph._get_action", mocked_get_action)
        action = self.heb_graph._get_any_action(range(5), None, None)
        check.is_none(action)

    def test_no_actions(self, mocker: MockerFixture):
        """should return 'Impossible' if no action is possible."""

        actions = ["Impossible", "Impossible", "Impossible", "Impossible", "Impossible"]

        _actions = deepcopy(actions)

        def mocked_get_action(*args):
            return _actions.pop(0)

        mocker.patch("hebg.heb_graph.HEBGraph._get_action", mocked_get_action)
        action = self.heb_graph._get_any_action(range(5), None, None)
        check.equal(action, "Impossible")

        action = self.heb_graph._get_any_action([], None, None)
        check.equal(action, "Impossible")

    @pytest.mark.parametrize("any_mode", HEBGraph.ANY_MODES)
    def test_any_mode_(self, any_mode, mocker: MockerFixture):
        actions = [0, "Impossible", 2, 3, "Impossible"]

        _actions = deepcopy(actions)

        def mocked_get_action(*args):
            return _actions.pop(0)

        expected_actions = {"first": (0,), "last": (3,), "random": (0, 2, 3)}

        mocker.patch("hebg.heb_graph.HEBGraph._get_action", mocked_get_action)
        self.heb_graph.any_mode = any_mode
        action = self.heb_graph._get_any_action(range(5), None, None)
        check.is_in(action, expected_actions[any_mode])


class TestHEBGraphGetAction:

    """HEBGraph._get_action"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""
        self.behavior = Behavior("behavior_name")
        self.heb_graph = HEBGraph(self.behavior)

    def test_action(self):
        """should return action given by an Action node."""
        expected_action = "action_action"

        class DummyAction:
            """DummyNode"""

            type = "action"

            def __call__(self, observation):
                return expected_action

        action_node = DummyAction()
        action = self.heb_graph._get_action(action_node, None, None)
        check.equal(action, expected_action)

    def test_empty(self, mocker: MockerFixture):
        """should return next successor returns when given an Empty node."""

        expected_action = "action_action"

        class DummyAction:
            """DummyNode"""

            type = "action"

            def __call__(self, observation):
                return expected_action

        mocker.patch(
            "hebg.heb_graph.HEBGraph.successors",
            lambda self, node: iter([DummyAction()]),
        )

        class DummyEmpty:
            """DummyNode"""

            type = "empty"

        empty_node = DummyEmpty()
        action = self.heb_graph._get_action(empty_node, None, None)
        check.equal(action, expected_action)

    def test_unknowed_node_type(self):
        """should raise ValueError if node.type is unknowed."""

        class DummyNode:
            """DummyNode"""

            type = "random_type_error"

        node = DummyNode()
        with pytest.raises(ValueError):
            self.heb_graph._get_action(node, None, None)

    def test_behavior_in_search(self):
        """should return 'Impossible' if behavior is already in search to avoid cycles."""

        class DummyBehavior:
            """DummyBehavior"""

            type = "behavior"
            name = "behavior_already_in_search"

            def __str__(self) -> str:
                return self.name

        node = DummyBehavior()
        action = self.heb_graph._get_action(node, None, ["behavior_already_in_search"])
        check.equal(action, "Impossible")

    def test_behavior_call(self):
        """should return behavior's return if behavior can be called."""
        expected_action = "behavior_action"

        class DummyBehavior:
            """DummyBehavior"""

            type = "behavior"
            name = "behavior_name"

            def __str__(self) -> str:
                return self.name

            def __call__(self, *args) -> str:
                return expected_action

        node = DummyBehavior()
        action = self.heb_graph._get_action(node, None, [])
        check.equal(action, expected_action)

    def test_behavior_by_all_behaviors(self):
        """should use all_behaviors if behavior cannot be called."""
        expected_action = "behavior_action"

        class DummyTrueBehavior:
            """DummyBehavior"""

            type = "behavior"
            name = "behavior_name"

            def __str__(self) -> str:
                return self.name

            def __call__(self, *args):
                return expected_action

        class DummyBehaviorInGraph:
            """DummyBehavior"""

            type = "behavior"
            name = "behavior_name"

            def __str__(self) -> str:
                return self.name

            def __call__(self, *args):
                raise NotImplementedError

        true_node = DummyTrueBehavior()
        self.heb_graph.all_behaviors = {true_node.name: true_node}

        node_in_graph = DummyBehaviorInGraph()
        action = self.heb_graph._get_action(
            node_in_graph,
            observation=None,
            behaviors_in_search=[],
        )
        check.equal(action, expected_action)

    def test_feature_condition(self, mocker: MockerFixture):
        """should use FeatureCondition's given index to orient in graph."""

        class DummyAction:
            """DummyAction"""

            type = "action"
            name = "action_name"

            def __init__(self, i) -> None:
                self.index = i
                self.action = f"action_{i}"

            def __str__(self) -> str:
                return self.name

            def __call__(self, *args) -> str:
                return self.action

        class DummyFeatureCondition:
            """DummyFeatureCondition"""

            type = "feature_condition"
            name = "feature_condition_name"

            def __init__(self, i) -> None:
                self.fc_index = i

            def __str__(self) -> str:
                return self.name

            def __call__(self, *args) -> int:
                return self.fc_index

        actions = [DummyAction(i) for i in range(3)]
        mocker.patch(
            "hebg.heb_graph.HEBGraph.successors",
            lambda self, node: actions,
        )

        class DummyEdges:
            """DummyEdges"""

            def __getitem__(self, e):
                return {"index": e[1].index}

        mocker.patch("hebg.heb_graph.HEBGraph.edges", DummyEdges())
        mocker.patch(
            "hebg.heb_graph.HEBGraph._get_any_action",
            lambda self, next_nodes, observation, behaviors_in_search: next_nodes[0](
                observation
            ),
        )

        for fc_index in range(3):
            node = DummyFeatureCondition(fc_index)
            action = self.heb_graph._get_action(node, None, [])
            expected_action = f"action_{fc_index}"
            check.equal(action, expected_action)

    def test_feature_condition_index_value_error(self, mocker: MockerFixture):
        """should raise ValueError if FeatureCondition's given index represents no successor."""

        class DummyAction:
            """DummyAction"""

            type = "action"
            name = "action_name"

            def __init__(self, i) -> None:
                self.index = i
                self.action = f"action_{i}"

            def __str__(self) -> str:
                return self.name

            def __call__(self, *args) -> str:
                return self.action

        class DummyFeatureCondition:
            """DummyFeatureCondition"""

            type = "feature_condition"
            name = "feature_condition_name"

            def __init__(self, i) -> None:
                self.fc_index = i

            def __str__(self) -> str:
                return self.name

            def __call__(self, *args) -> int:
                return self.fc_index

        actions = [DummyAction(i) for i in range(3)]
        mocker.patch(
            "hebg.heb_graph.HEBGraph.successors",
            lambda self, node: actions,
        )

        class DummyEdges:
            """DummyEdges"""

            def __getitem__(self, e):
                return {"index": e[1].index}

        mocker.patch("hebg.heb_graph.HEBGraph.edges", DummyEdges())

        node = DummyFeatureCondition(4)
        with pytest.raises(ValueError):
            self.heb_graph._get_action(node, None, [])



"""Module to unroll HEBGraph.

Unrolling means expanding each sub-behavior node as it's own graph in the global HEBGraph.
Behaviors that do not have a graph (Unexplainable behaviors) should stay as is in the graph.

"""
from copy import copy
from typing import TYPE_CHECKING, Dict

from networkx import relabel_nodes

from hebg.node import Node
from hebg.behavior import Behavior
from hebg.graph import compute_levels

BEHAVIOR_SEPARATOR = ">"

if TYPE_CHECKING:
    from hebg.heb_graph import HEBGraph


def unroll_graph(graph: "HEBGraph", add_prefix=True) -> "HEBGraph":
    """Build the the unrolled HEBGraph.

    The HEBGraph as the same behavior but every behavior node is recursively replaced
    by it's own HEBGraph if it can be computed.

    Returns:
        This HEBGraph's unrolled HEBGraph.

    """

    unrolled_graph: "HEBGraph" = copy(graph)
    for node in unrolled_graph.nodes():
        if not isinstance(node, Behavior):
            continue
        unrolled_graph = unroll_behavior(unrolled_graph, node, add_prefix)

    compute_levels(unrolled_graph)
    return unrolled_graph


def unroll_behavior(
    graph: "HEBGraph", behavior: Behavior, add_prefix: bool
) -> "HEBGraph":
    """Unroll a behavior node in a given HEBGraph

    Args:
        graph (HEBGraph): HEBGraph to unroll the behavior in.
        behavior (Behavior): Behavior node to unroll, must be in the given graph.
        add_prefix (bool): If True, adds a name prefix to keep nodes different.

    Returns:
        HEBGraph: Initial graph with unrolled behavior.
    """
    # Look for name reference.
    if str(behavior) in graph.all_behaviors:
        behavior = graph.all_behaviors[behavior.name]
    try:
        node_graph = unroll_graph(behavior.graph, add_prefix=add_prefix)
    except NotImplementedError:
        # If we cannot unroll, we keep it as is
        return graph

    # Relabel graph nodes to obtain disjoint node labels (if more that one node).
    if add_prefix and len(node_graph.nodes()) > 1:
        add_prefix_to_graph(node_graph, behavior.name + BEHAVIOR_SEPARATOR)

    # Replace the behavior node by the unrolled behavior's graph
    graph = compose_heb_graphs(graph, node_graph)
    for edge_u, _, data in graph.in_edges(behavior, data=True):
        for root in node_graph.roots:
            graph.add_edge(edge_u, root, **data)
    for _, edge_v, data in graph.out_edges(behavior):
        for root in node_graph.roots:
            graph.add_edge(root, edge_v, **data)

    graph.remove_node(behavior)
    return graph


def add_prefix_to_graph(graph: "HEBGraph", prefix: str) -> None:
    """Rename graph to obtain disjoint node labels."""
    if prefix is None:
        return graph

    def rename(node: Node):
        new_node = copy(node)
        new_node.name = prefix + node.name
        return new_node

    return relabel_nodes(graph, rename, copy=False)


def group_behaviors_points(
    pos: Dict[Node, tuple],
    graph: "HEBGraph",
) -> Dict[tuple, list]:
    """Group nodes positions of an HEBGraph by sub-behavior.

    Args:
        pos (Dict[Node, tuple]): Positions of nodes.
        graph (HEBGraph): Graph.

    Returns:
        Dict[tuple, list]: A dictionary of nodes grouped by their behavior's hierarchy.
    """
    points_grouped_by_behavior: Dict[tuple, list] = {}
    for node in graph.nodes():
        groups = str(node).split(BEHAVIOR_SEPARATOR)
        if len(groups) > 1:
            for i in range(len(groups[:-1])):
                key = tuple(groups[: -1 - i])
                point = pos[node]
                try:
                    points_grouped_by_behavior[key].append(point)
                except KeyError:
                    points_grouped_by_behavior[key] = [point]
    return points_grouped_by_behavior


def compose_heb_graphs(graph_of_reference: "HEBGraph", other_graph: "HEBGraph"):
    """Returns a new_graph of graph_of_reference composed with other_graph.

    Composition is the simple union of the node sets and edge sets.
    The node sets of the graph_of_reference and other_graph do not need to be disjoint.

    Args:
        graph_of_reference, other_graph : HEBGraphs to compose.

    Returns:
        A new HEBGraph with the same type as graph_of_reference.

    """
    new_graph = graph_of_reference.__class__(
        graph_of_reference.behavior,
        all_behaviors=graph_of_reference.all_behaviors,
        any_mode=graph_of_reference.any_mode,
    )
    # add graph attributes, H attributes take precedent over G attributes
    new_graph.graph.update(graph_of_reference.graph)
    new_graph.graph.update(other_graph.graph)

    new_graph.add_nodes_from(graph_of_reference.nodes(data=True))
    new_graph.add_nodes_from(other_graph.nodes(data=True))

    if graph_of_reference.is_multigraph():
        new_graph.add_edges_from(graph_of_reference.edges(keys=True, data=True))
    else:
        new_graph.add_edges_from(graph_of_reference.edges(data=True))
    if other_graph.is_multigraph():
        new_graph.add_edges_from(other_graph.edges(keys=True, data=True))
    else:
        new_graph.add_edges_from(other_graph.edges(data=True))
    return new_graph



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Module containing layouts to draw graphs. """

from hebg.layouts.deterministic import staircase_layout
from hebg.layouts.metabased import leveled_layout_energy



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>
# pylint: disable=protected-access

""" Deterministic layouts """

import networkx as nx
import numpy as np

from hebg.graph import get_roots


def staircase_layout(graph: nx.DiGraph, center=None):
    """Compute specific default positions for an DiGraph.

    Requires graph to have a 'nodes_by_level' attribute.

    Args:
        graph: A networkx DiGraph with 'nodes_by_level' attribute (see compute_levels).
        center (Optional): Center of the graph layout.

    Returns:
        pos: Positions of each node.

    """

    def place_successors(pos, pos_by_level, node, level) -> int:
        if not level in pos_by_level:
            pos_by_level[level] = pos_by_level[level - 1]
        pos_by_level[level] = max(pos[node][0], pos_by_level[level])
        succs = list(graph.successors(node))
        if len(succs) == 0:
            return 1
        succs_order = np.argsort([graph.edges[node, succ]["index"] for succ in succs])
        for index, succ_id in enumerate(succs_order):
            succ = succs[succ_id]
            if succ in pos:
                continue
            pos[succ] = [pos_by_level[level], -level]
            if index == 0:
                pos[node][0] = max(pos[node][0], pos[succ][0])
                pos_by_level[level - 1] = max(pos_by_level[level - 1], pos[node][0])
            n_succs = place_successors(pos, pos_by_level, succ, level + 1)
            pos_by_level[level] += n_succs
        return len(succs)

    graph, _ = nx.drawing.layout._process_params(graph, center, dim=2)
    pos = {}
    pos_by_level = {0: 0}
    for node in get_roots(graph):
        pos[node] = [pos_by_level[0], 0]
        pos_by_level[0] += place_successors(pos, pos_by_level, node, 1)
    return pos



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Integration tests for the hebg.metrics.complexity.complexities module. """

import pytest
import pytest_check as check

from hebg.metrics.complexity.complexities import learning_complexity
from hebg import Action, Behavior, FeatureCondition


class TestComplexities:
    """Complexities"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Unit tests for the hebg.node module. """

import pytest
import pytest_check as check

from hebg.node import Node, Action, FeatureCondition, EmptyNode


class TestNode:

    """Node"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""

    def test_node_type(self):
        """should have correct node_type and raise ValueError otherwise."""
        with pytest.raises(ValueError):
            Node("", "")
        for node_type in ("action", "feature_condition", "behavior", "empty"):
            node = Node("", node_type)
            check.equal(node.type, node_type)

    def test_node_name(self):
        """should have name as attribute."""
        name = "node_name"
        node = Node(name, "empty")
        check.equal(node.name, name)

    def test_node_call(self):
        """should raise NotImplementedError on call."""
        node = Node("", "empty")
        with pytest.raises(NotImplementedError):
            node(None)


class TestAction:

    """Action"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""

    def test_node_type(self):
        """should have 'action' as node_type."""
        node = Action("", "")
        check.equal(node.type, "action")

    def test_node_call(self):
        """should return Action.action when called."""
        action = "action_action"
        node = Action(action, "action_name")
        check.equal(node(None), action)
        check.equal(node.action, action)


class TestFeatureCondition:

    """FeatureCondition"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""

    def test_node_type(self):
        """should have 'feature_condition' as node_type."""
        node = FeatureCondition("")
        check.equal(node.type, "feature_condition")

    def test_node_call(self):
        """should raise NotImplementedError on call."""
        node = FeatureCondition("")
        with pytest.raises(NotImplementedError):
            node(None)


class TestEmptyNode:

    """EmptyNode"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""

    def test_node_type(self):
        """should have 'empty' as node_type."""
        node = EmptyNode("")
        check.equal(node.type, "empty")

    def test_node_call(self):
        """should return 1 when called."""
        node = EmptyNode("")
        check.equal(node(None), 1)



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Module for complexity computation methods. """

from hebg.metrics.complexity.complexities import general_complexity, learning_complexity



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Integration tests for the initial paper examples. """

from typing import Dict, List
from copy import deepcopy

# import matplotlib.pyplot as plt


import pytest
import pytest_check as check

from itertools import permutations
from networkx.classes.digraph import DiGraph
from networkx import is_isomorphic

from hebg import Action, Behavior, FeatureCondition, HEBGraph
from hebg.metrics.histograms import behaviors_histograms, cumulated_hebgraph_histogram
from hebg.metrics.complexity.complexities import learning_complexity
from hebg.requirements_graph import build_requirement_graph
from hebg.unrolling import BEHAVIOR_SEPARATOR

from tests.examples.behaviors.report_example import Behavior0, Behavior1, Behavior2


class TestPaperBasicExamples:
    """Basic examples from the initial paper"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""

        self.actions: List[Action] = [Action(i, complexity=1) for i in range(3)]
        self.feature_conditions: List[FeatureCondition] = [
            FeatureCondition(f"feature {i}", complexity=1) for i in range(6)
        ]
        self.behaviors: List[Behavior] = [Behavior0(), Behavior1(), Behavior2()]

        self.expected_behavior_histograms: Dict[Behavior, Dict[Action, int]] = {
            self.behaviors[0]: {
                self.actions[0]: 1,
                self.actions[1]: 1,
                self.feature_conditions[0]: 1,
            },
            self.behaviors[1]: {
                self.actions[0]: 1,
                self.actions[2]: 1,
                self.behaviors[0]: 1,
                self.feature_conditions[1]: 1,
                self.feature_conditions[2]: 1,
            },
            self.behaviors[2]: {
                self.actions[0]: 1,
                self.behaviors[0]: 1,
                self.behaviors[1]: 2,
                self.feature_conditions[3]: 1,
                self.feature_conditions[4]: 1,
                self.feature_conditions[5]: 1,
            },
        }

    def test_histograms(self):
        """should give expected histograms."""
        check.equal(
            behaviors_histograms(self.behaviors), self.expected_behavior_histograms
        )

    def test_cumulated_histograms(self):
        """should give expected cumulated histograms."""
        expected_cumulated_histograms = {
            self.behaviors[0]: {
                self.actions[0]: 1,
                self.actions[1]: 1,
                self.feature_conditions[0]: 1,
            },
            self.behaviors[1]: {
                self.actions[0]: 2,
                self.actions[2]: 1,
                self.actions[1]: 1,
                self.feature_conditions[0]: 1,
                self.feature_conditions[1]: 1,
                self.feature_conditions[2]: 1,
                self.behaviors[0]: 1,
            },
            self.behaviors[2]: {
                self.actions[0]: 6,
                self.actions[1]: 3,
                self.actions[2]: 2,
                self.feature_conditions[0]: 3,
                self.feature_conditions[1]: 2,
                self.feature_conditions[2]: 2,
                self.feature_conditions[3]: 1,
                self.feature_conditions[4]: 1,
                self.feature_conditions[5]: 1,
                self.behaviors[0]: 3,
                self.behaviors[1]: 2,
            },
        }
        for behavior in self.behaviors:
            check.equal(
                cumulated_hebgraph_histogram(behavior.graph),
                expected_cumulated_histograms[behavior],
            )

    def test_learning_complexity(self):
        """should give expected learning_complexity."""
        expected_learning_complexities = {
            self.behaviors[0]: 3,
            self.behaviors[1]: 6,
            self.behaviors[2]: 9,
        }
        expected_saved_complexities = {
            self.behaviors[0]: 0,
            self.behaviors[1]: 1,
            self.behaviors[2]: 12,
        }

        for behavior in self.behaviors:
            c_learning, saved_complexity = learning_complexity(
                behavior, used_nodes_all=self.expected_behavior_histograms
            )

            print(
                f"{behavior}: {c_learning}|{expected_learning_complexities[behavior]}"
                f" {saved_complexity}|{expected_saved_complexities[behavior]}"
            )
            diff_complexity = abs(c_learning - expected_learning_complexities[behavior])
            diff_saved = abs(saved_complexity - expected_saved_complexities[behavior])
            check.less(diff_complexity, 1e-14)
            check.less(diff_saved, 1e-14)

    def test_codegen(self):
        expected_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class Behavior0(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['feature 0'](observation)",
                "        if edge_index == 0:",
                "            return self.actions['action 0'](observation)",
                "        if edge_index == 1:",
                "            return self.actions['action 1'](observation)",
                "class Behavior1(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['feature 1'](observation)",
                "        if edge_index == 0:",
                "            return self.known_behaviors['behavior 0'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['feature 2'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 0'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.actions['action 2'](observation)",
                "class Behavior2(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['feature 3'](observation)",
                "        if edge_index == 0:",
                "            edge_index_1 = self.feature_conditions['feature 4'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 0'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.known_behaviors['behavior 1'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['feature 5'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.known_behaviors['behavior 1'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.known_behaviors['behavior 0'](observation)",
                "BEHAVIOR_TO_NAME = {",
                "    'behavior 0': Behavior0,",
                "    'behavior 1': Behavior1,",
                "}",
            )
        )
        generated_code = self.behaviors[2].graph.generate_source_code()
        check.equal(generated_code, expected_code)

    def test_requirement_graph_edges(self):
        """should give expected requirement_graph edges."""
        expected_requirement_graph = DiGraph()
        for behavior in self.behaviors:
            expected_requirement_graph.add_node(behavior)
        expected_requirement_graph.add_edge(self.behaviors[0], self.behaviors[1])
        expected_requirement_graph.add_edge(self.behaviors[0], self.behaviors[2])
        expected_requirement_graph.add_edge(self.behaviors[1], self.behaviors[2])

        requirements_graph = build_requirement_graph(self.behaviors)
        for behavior, other_behavior in permutations(self.behaviors, 2):
            print(behavior, other_behavior)
            req_has_edge = requirements_graph.has_edge(behavior, other_behavior)
            expected_req_has_edge = expected_requirement_graph.has_edge(
                behavior, other_behavior
            )
            check.equal(req_has_edge, expected_req_has_edge)

    def test_requirement_graph_levels(self):
        """should give expected requirement_graph node levels (requirement depth)."""
        expected_levels = {
            self.behaviors[0]: 0,
            self.behaviors[1]: 1,
            self.behaviors[2]: 2,
        }
        requirements_graph = build_requirement_graph(self.behaviors)
        for behavior, level in requirements_graph.nodes(data="level"):
            check.equal(level, expected_levels[behavior])

    def test_unrolled_behaviors_graphs(self):
        """should give expected unrolled_behaviors_graphs for each example behaviors."""

        def lname(*args):
            return BEHAVIOR_SEPARATOR.join([str(arg) for arg in args])

        expected_graph_0 = deepcopy(self.behaviors[0].graph)

        expected_graph_1 = HEBGraph(self.behaviors[1])
        feature_0 = FeatureCondition(lname(self.behaviors[0], "feature 0"))
        expected_graph_1.add_edge(
            feature_0, Action(0, lname(self.behaviors[0], "action 0")), index=False
        )
        expected_graph_1.add_edge(
            feature_0, Action(1, lname(self.behaviors[0], "action 1")), index=True
        )
        feature_1 = FeatureCondition("feature 1")
        feature_2 = FeatureCondition("feature 2")
        expected_graph_1.add_edge(feature_1, feature_0, index=False)
        expected_graph_1.add_edge(feature_1, feature_2, index=True)
        expected_graph_1.add_edge(feature_2, Action(0), index=False)
        expected_graph_1.add_edge(feature_2, Action(2), index=True)

        expected_graph_2 = HEBGraph(self.behaviors[2])
        feature_3 = FeatureCondition("feature 3")
        feature_4 = FeatureCondition("feature 4")
        feature_5 = FeatureCondition("feature 5")
        expected_graph_2.add_edge(feature_3, feature_4, index=False)
        expected_graph_2.add_edge(feature_3, feature_5, index=True)
        expected_graph_2.add_edge(feature_4, Action(0), index=False)

        feature_0 = FeatureCondition(
            lname(self.behaviors[1], self.behaviors[0], "feature 0")
        )
        expected_graph_2.add_edge(
            feature_0,
            Action(0, lname(self.behaviors[1], self.behaviors[0], "action 0")),
            index=False,
        )
        expected_graph_2.add_edge(
            feature_0,
            Action(1, lname(self.behaviors[1], self.behaviors[0], "action 1")),
            index=True,
        )
        feature_1 = FeatureCondition(lname(self.behaviors[1], "feature 1"))
        feature_2 = FeatureCondition(lname(self.behaviors[1], "feature 2"))
        expected_graph_2.add_edge(feature_1, feature_0, index=False)
        expected_graph_2.add_edge(feature_1, feature_2, index=True)
        expected_graph_2.add_edge(
            feature_2, Action(0, lname(self.behaviors[1], "action 0")), index=False
        )
        expected_graph_2.add_edge(
            feature_2, Action(2, lname(self.behaviors[1], "action 2")), index=True
        )

        expected_graph_2.add_edge(feature_4, feature_1, index=True)

        feature_0_0 = FeatureCondition(lname(self.behaviors[0], "feature 0"))
        expected_graph_2.add_edge(
            feature_0_0,
            Action(0, lname(self.behaviors[0], "action 0")),
            index=False,
        )
        expected_graph_2.add_edge(
            feature_0_0,
            Action(1, lname(self.behaviors[0], "action 1")),
            index=True,
        )

        expected_graph_2.add_edge(feature_5, feature_1, index=False)
        expected_graph_2.add_edge(feature_5, feature_0_0, index=True)

        expected_graph = {
            self.behaviors[0]: expected_graph_0,
            self.behaviors[1]: expected_graph_1,
            self.behaviors[2]: expected_graph_2,
        }
        for behavior in self.behaviors:
            unrolled_graph = behavior.graph.unrolled_graph
            check.is_true(is_isomorphic(unrolled_graph, expected_graph[behavior]))

            # fig, axes = plt.subplots(1, 2)
            # unrolled_graph = behavior.graph.unrolled_graph
            # unrolled_graph.draw(axes[0], draw_behaviors_hulls=True)
            # expected_graph[behavior].draw(axes[1], draw_behaviors_hulls=True)
            # plt.show()



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Behavior of HEBGraphs when called. """

import pytest_check as check

from hebg.node import Action
from tests.examples.behaviors import (
    FundamentalBehavior,
    AA_Behavior,
    F_A_Behavior,
    F_F_A_Behavior,
    AF_A_Behavior,
    F_AA_Behavior,
)
from tests.examples.feature_conditions import ThresholdFeatureCondition


def test_a_graph():
    """(A) Fundamental behaviors (single action) should work properly."""
    action_id = 42
    behavior = FundamentalBehavior(Action(action_id))
    check.equal(behavior(None), action_id)


def test_f_a_graph():
    """(F-A) Feature condition should orient path properly."""
    feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)
    actions = {0: Action(0), 1: Action(1)}
    behavior = F_A_Behavior("F_A", feature_condition, actions)
    check.equal(behavior(1), 1)
    check.equal(behavior(-1), 0)


def test_f_f_a_graph():
    """(F-F-A) Feature condition should orient path properly in double chain."""
    behavior = F_F_A_Behavior("F_F_A")
    check.equal(behavior(-2), 0)
    check.equal(behavior(-1), 1)
    check.equal(behavior(1), 2)
    check.equal(behavior(2), 3)


def test_aa_graph():
    """(AA) Should choose between roots depending on 'any_mode'."""
    behavior = AA_Behavior("AA", any_mode="first")
    check.equal(behavior(None), 0)

    behavior = AA_Behavior("AA", any_mode="last")
    check.equal(behavior(None), 1)


def test_af_a_graph():
    """(AF-A) Should choose between roots depending on 'any_mode'."""
    behavior = AF_A_Behavior("AF_A", any_mode="first")
    check.equal(behavior(1), 0)
    check.equal(behavior(-1), 0)

    behavior = AF_A_Behavior("AF_A", any_mode="last")
    check.equal(behavior(1), 1)
    check.equal(behavior(-1), 2)


def test_f_af_a_graph():
    """(F-AA) Should choose between condition edges depending on 'any_mode'."""
    behavior = F_AA_Behavior("F_AA", any_mode="first")
    check.equal(behavior(1), 0)
    check.equal(behavior(-1), 1)

    behavior = F_AA_Behavior("F_AA", any_mode="last")
    check.equal(behavior(1), 0)
    check.equal(behavior(-1), 2)



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" HEBGraph used nodes histograms computation. """

from typing import TYPE_CHECKING, Dict, List, Tuple
from warnings import warn

import numpy as np


from hebg.metrics.complexity.utils import update_sum_dict
from hebg.behavior import Behavior
from hebg.node import Action, FeatureCondition

if TYPE_CHECKING:
    from hebg import HEBGraph, Node


def behaviors_histograms(
    behaviors: List["Behavior"],
    default_node_complexity: float = 1.0,
) -> Dict["Behavior", Dict["Node", int]]:
    """Compute the used nodes histograms for a list of Behavior.

    Args:
        behaviors: List of Behavior to compute histograms of.
        default_node_complexity: Default node complexity if Node has no attribute complexity.

    Return:
        Dictionary of dictionaries of the number of use for each used node, for each behavior.
    """
    return behaviors_histograms_and_complexites(behaviors, default_node_complexity)[0]


def behaviors_histograms_and_complexites(
    behaviors: List["Behavior"],
    default_node_complexity: float = 1.0,
) -> Tuple[Dict["Behavior", Dict["Node", int]], Dict["Behavior", float]]:
    """Compute the used nodes histograms for a list of Behavior.

    Args:
        behaviors: List of Behavior to compute histograms of.
        default_node_complexity: Default node complexity if Node has no attribute complexity.

    Return:
        Tuple of two elements:
        - Dictionary of dictionaries of the number of use for each used node, for each behavior.
        - Dictionary computed complexity for each behavior.
    """
    histograms: Dict["Behavior", Dict["Node", int]] = {}
    complexities: Dict["Behavior", float] = {}
    for behavior in behaviors:
        try:
            graph = behavior.graph
        except NotImplementedError:
            warn(
                f"Could not load graph for behavior: {behavior}."
                "Skipping histogram computation."
            )
            continue
        histogram, complexity = hebgraph_histogram_and_complexity(
            graph,
            default_node_complexity=default_node_complexity,
        )
        histograms[behavior] = histogram
        complexities[behavior] = complexity
    return histograms, complexities


def hebgraph_histogram_and_complexity(
    graph: "HEBGraph", default_node_complexity: float = 1.0
) -> Tuple[Dict["Node", int], float]:
    """Compute the used nodes histogram for a Behavior.

    Args:
        behavior: Behavior to compute histogram of.
        default_node_complexity: Default node complexity if Node has no attribute complexity.

    Return:
        Tuple composed of two element:
        - Dictionary of the number of use for each used node in the graph.
        - Computed total complexity of the graph.

    """
    nodes_histograms, nodes_complexities = nodes_histograms_and_complexities(
        graph, default_node_complexity
    )
    root = graph.graph["nodes_by_level"][0][0]  # Assumes a single root
    return nodes_histograms[root], nodes_complexities[root]


def cumulated_hebgraph_histogram(
    graph: "HEBGraph", default_node_complexity: float = 1.0
) -> Dict["Node", int]:
    """Unroll the hebgraph histogram by accumulating sub-behaviors histograms.

    Args:
        graph (HEBGraph): The HEBgraph to compute the cumulated histogram of.
        default_node_complexity (float, optional): Default node complexity. Defaults to 1.0.

    Returns:
        Dict[Node, int]: Cumulated histogram of used nodes with unrolled behaviors.
    """
    histogram, _ = hebgraph_histogram_and_complexity(graph, default_node_complexity)
    histograms = {graph.behavior: histogram}
    sub_behaviors = [
        node
        for node in histogram
        if isinstance(node, Behavior) and node != graph.behavior
    ]
    for i, behavior in enumerate(sub_behaviors):
        if behavior.name in graph.all_behaviors:
            sub_behaviors[i] = graph.all_behaviors[behavior.name]

    histograms.update(behaviors_histograms(sub_behaviors, default_node_complexity))

    done = False
    behavior_iteration = {}  # Stoping condition
    while not done:
        behaviors = [node for node in histogram if isinstance(node, Behavior)]
        behavior_iteration, histogram, histograms = _iterate_cumulation(
            graph=graph,
            behaviors=behaviors,
            histogram=histogram,
            histograms=histograms,
            behavior_iteration=behavior_iteration,
            default_node_complexity=default_node_complexity,
        )

        # Recompute behaviors because some might have been added
        behaviors = [node for node in histogram if isinstance(node, Behavior)]

        # We stop when all behaviors have been iterated on enough.
        if all(behavior in behavior_iteration for behavior in behaviors):
            done = all(
                behavior_iteration[behavior] == histogram[behavior]
                for behavior in behaviors
            )

    return histogram


def _iterate_cumulation(
    graph: "HEBGraph",
    behaviors: List["Behavior"],
    histogram: Dict["Node", int],
    histograms: Dict["Behavior", Dict["Node", int]],
    behavior_iteration: Dict["Behavior", int],
    default_node_complexity: float = 1.0,
):
    """Iterate once on every behavior accumulating histograms.

    Args:
        graph (HEBGraph): Current graph being iterated on.
        behaviors (List[Behavior]): List of behaviors to iterate.
        histogram (Dict[Node, int]): Histogram being accumulated.
        histograms (Dict[Behavior, Dict[Node, int]]): Histograms of knowed behaviors.
        behavior_iteration (Dict[Behavior, int]): Number of iterations already done
            for each behavior.

    Returns:
        Tuple of three updated dictionaries. (behavior_iteration, histogram, histograms).
    """
    for behavior in behaviors:
        already_iterated = (
            behavior_iteration[behavior] if behavior in behavior_iteration else 0
        )
        n_used = max(0, histogram[behavior] - already_iterated)
        if behavior not in histograms:
            if behavior.name in graph.all_behaviors:
                behavior = graph.all_behaviors[behavior.name]
            try:
                behavior_graph = behavior.graph
            except NotImplementedError:
                if behavior not in behavior_iteration:
                    behavior_iteration[behavior] = 0
                behavior_iteration[behavior] += 1
                continue
            sub_histogram, _ = hebgraph_histogram_and_complexity(
                behavior_graph, default_node_complexity
            )
            histograms[behavior] = sub_histogram
        for _ in range(n_used):
            if behavior in histograms[behavior]:
                histograms[behavior].pop(behavior)
            histogram = update_sum_dict(histogram, histograms[behavior])
            if behavior not in behavior_iteration:
                behavior_iteration[behavior] = 0
            behavior_iteration[behavior] += 1
    return behavior_iteration, histogram, histograms


def nodes_histograms_and_complexities(
    graph: "HEBGraph",
    default_node_complexity: float = 1.0,
    _behaviors_in_search=None,
):
    """Compute the number of times each node of the graph is present
    while computing complexities to find the least complex path.

    Args:
        graph (HEBGraph): HEBGraph to compute the histogram of.
        default_node_complexity (float, optional): Default node complexity if undefined.
            Defaults to 1.0.
        _behaviors_in_search (List[Behavior], optional): List of behaviors already in reccursive
            pile to avoid infinite cycle. Defaults to None.

    Returns:
        Tuple[Dict[Node, Dict[Node, int]], Dict[Node, float]]: Tuple of two values:
        - Dictionary of subnode used for each node. (nodes_used_nodes)
        - Dictionary of computed smallest path complexities for each node. (complexities)
    """
    nodes_by_level = graph.graph["nodes_by_level"]
    depth = graph.graph["depth"]

    _behaviors_in_search = [] if _behaviors_in_search is None else _behaviors_in_search
    _behaviors_in_search.append(str(graph.behavior))

    complexities = {}
    nodes_used_nodes = {}
    for node in graph.nodes:
        if isinstance(node, (Action, FeatureCondition)):
            complexities[node] = node.complexity
            nodes_used_nodes[node] = {node: 1}

    for level in range(depth + 1)[::-1]:
        for node in nodes_by_level[level]:
            node_complexity = 0
            node_used_nodes = {}

            # Best successors accumulated histograms and complexity
            succ_by_index, complexities_by_index = _successors_by_index(
                graph, node, complexities
            )
            for index, values in complexities_by_index.items():
                min_index = np.argmin(values)
                choosen_succ = succ_by_index[index][min_index]
                node_used_nodes = update_sum_dict(
                    node_used_nodes, nodes_used_nodes[choosen_succ]
                )
                node_complexity += values[min_index]

            # Node only histogram and complexity
            (
                node_only_used_behaviors,
                node_only_complexity,
            ) = _get_node_histogram_complexity(
                node,
                default_node_complexity=default_node_complexity,
                behaviors_in_search=_behaviors_in_search,
            )
            node_used_nodes = update_sum_dict(node_used_nodes, node_only_used_behaviors)
            node_complexity += node_only_complexity

            complexities[node] = node_complexity
            nodes_used_nodes[node] = node_used_nodes
    return nodes_used_nodes, complexities


def _successors_by_index(
    graph: "HEBGraph", node: "Node", complexities: Dict["Node", float]
) -> Tuple[Dict[int, List["Node"]], Dict[int, List[float]]]:
    """Group successors and their complexities by index.

    Args:
        graph: The HEBGraph to use.
        node: The Node from which we want to group successors.
        complexities: Dictionary of complexities for each potential successor node.

    Return:
        Tuple composed of a dictionary of successors for each index
            and a dictionary of complexities for each index.

    """
    complexities_by_index = {}
    succ_by_index = {}
    for succ in graph.successors(node):
        succ_complexity = complexities[succ]
        index = int(graph.edges[node, succ]["index"])
        if index not in complexities_by_index:
            complexities_by_index[index] = []
        if index not in succ_by_index:
            succ_by_index[index] = []
        complexities_by_index[index].append(succ_complexity)
        succ_by_index[index].append(succ)
    return succ_by_index, complexities_by_index


def _get_node_histogram_complexity(
    node: "Node", behaviors_in_search=None, default_node_complexity: float = 1.0
) -> Tuple[Dict["Node", int], float]:
    """Compute the used nodes histogram and complexity of a single node.

    Args:
        node: The Node from which we want to compute the complexity.
        behaviors_in_search: List of Behavior already in search to avoid circular search.
        default_node_complexity: Default node complexity if Node has no attribute complexity.

    Return:
        Tuple composed of a dictionary of the number of use for each used Node by the given node
            and the given node complexity.

    """

    if node.type == "behavior":
        if behaviors_in_search is not None and str(node) in behaviors_in_search:
            return {}, np.inf
    if node.type in ("action", "feature_condition", "behavior"):
        try:
            node_complexity = node.complexity
        except AttributeError:
            node_complexity = default_node_complexity
        return {node: 1}, node_complexity
    if node.type == "empty":
        return {}, 0
    raise ValueError(f"Unkowned node type {node.type}")



"""Module for code generation from HEBGraph."""

from re import sub

from typing import TYPE_CHECKING, List, Dict, Set, Tuple

from hebg.node import Node, Action, FeatureCondition
from hebg.behavior import Behavior
from hebg.unrolling import BEHAVIOR_SEPARATOR
from hebg.graph import get_roots, get_successors_with_index
from hebg.metrics.histograms import cumulated_hebgraph_histogram

if TYPE_CHECKING:
    from hebg.heb_graph import HEBGraph


class GeneratedBehavior:
    """Base class for generated behaviors.

    Used to reduce the overhead of abstracting a behavior."""

    def __init__(
        self,
        actions: Dict[str, "Action"] = None,
        feature_conditions: Dict[str, "FeatureCondition"] = None,
        behaviors: Dict[str, "Behavior"] = None,
    ):
        self.actions = actions if actions is not None else {}
        self.feature_conditions = feature_conditions if actions is not None else {}
        self.known_behaviors = behaviors if behaviors is not None else {}


def get_hebg_source(graph: "HEBGraph") -> str:
    """Builds the generated source code corresponding to the HEBGraph behavior.

    Args:
        graph (HEBGraph): HEBGraph to generate the source code from.

    Returns:
        str: Python source code corresponding the behavior of the HEBGraph.
    """
    class_codelines = get_behavior_class_codelines(graph)
    source = "\n".join(class_codelines)
    return source


def get_behavior_class_codelines(
    graph: "HEBGraph",
    behaviors_histogram: Dict["Behavior", Dict["Node", int]] = None,
    add_dependencies: bool = True,
) -> List[str]:
    """Generate codelines of the whole GeneratedBehavior from a HEBGraph.

    Args:
        graph (HEBGraph): HEBGraph to generate the behavior from.
        behaviors_histogram (Dict[Behavior, Dict[Node, int]], optional): Histogram of uses
            of all behaviors needed for the given graph. Defaults to None.
        add_dependencies (bool, optional): If True, codelines will include other GeneratedBehavior
            from sub-behaviors of the given HEBGraph and a hashmap to map behavior name
            to coresponding GeneratedBehavior. Defaults to True.

    Returns:
        List[str]: Codelines generated of the HEBGraph GeneratedBehavior.
    """
    if behaviors_histogram is None:
        behaviors_histogram = cumulated_hebgraph_histogram(graph)

    (
        dependencies_codelines,
        behaviors_incall_codelines,
        hashmap_codelines,
    ) = get_dependencies_codelines(
        graph,
        behaviors_histogram,
        add_dependencies_codelines=add_dependencies,
    )

    class_codelines = []
    # Other behaviors dependencies
    if add_dependencies:
        class_codelines += ["from hebg.codegen import GeneratedBehavior", ""]
        for _behavior, dependency_codelines in dependencies_codelines.items():
            class_codelines += dependency_codelines

    # Class overhead
    behavior_class_name = _to_camel_case(graph.behavior.name.capitalize())
    class_codelines.append(f"class {behavior_class_name}(GeneratedBehavior):")
    # Call function
    class_codelines += get_behavior_call_codelines(
        graph,
        behaviors_incall_codelines=behaviors_incall_codelines,
    )
    # Dependencies hashmap
    if add_dependencies:
        class_codelines += hashmap_codelines
    return class_codelines


def get_dependencies_codelines(
    graph: "HEBGraph",
    behaviors_histogram: Dict["Behavior", Dict["Node", int]],
    add_dependencies_codelines: bool = True,
) -> Tuple[Dict["Behavior", List[str]], Set["Behavior"], Dict[str, str]]:
    """Parse dependencies of the given HEBGraph behavior's.

    Args:
        graph (HEBGraph): HEBGraph to parse dependecies from.
        behaviors_histogram (Dict[Behavior, Dict[Node, int]]): _description_
        add_dependencies_codelines (bool, optional): _description_. Defaults to True.

    Returns:
        Tuple of three elements:
        - dependencies_codelines (Dict[Behavior, List[str]]): Codelines of the GeneratedBehavior for
            each of the behavior used in the HEBGraph if they can be computed, else a comment.
        - behaviors_incall_codelines (Set[Behavior]): Set of behaviors that should be directly
            unrolled in the call function (because the abstraction is not worth it).
        - hashmap_codelines (List[str]): Codelines of the map between behavior names
            and coresponding GeneratedBehavior.
    """
    dependencies_codelines: Dict["Behavior", List[str]] = {}
    behaviors_incall_codelines: Set["Behavior"] = set()
    dependencies_hashmap: Dict[str, str] = {}
    for behavior, n_used in sorted(
        behaviors_histogram.items(), key=lambda x: x[1], reverse=True
    ):
        if not isinstance(behavior, Behavior):
            continue
        if behavior.name in graph.all_behaviors:
            behavior = graph.all_behaviors[behavior.name]

        try:
            sub_graph = behavior.graph
        except NotImplementedError:
            # If subgraph cannot be computed, we simply have a ref
            if add_dependencies_codelines:
                dependencies_codelines[behavior] = [
                    f"# Require '{behavior.name}' behavior to be given."
                ]
            continue

        if n_used == 1 or len(sub_graph.nodes()) == 1:
            dependencies_codelines[behavior] = []
            behaviors_incall_codelines.add(behavior)
            continue

        if add_dependencies_codelines:
            dependencies_codelines[behavior] = get_behavior_class_codelines(
                sub_graph, behaviors_histogram, add_dependencies=False
            )
            dependencies_hashmap[behavior.name] = _to_camel_case(
                behavior.name.capitalize()
            )

    hashmap_codelines = []
    if dependencies_hashmap:
        hashmap_codelines = ["BEHAVIOR_TO_NAME = {"]
        hashmap_codelines += [
            f"    '{name}': {class_name},"
            for name, class_name in dependencies_hashmap.items()
        ]
        hashmap_codelines += ["}"]

    return dependencies_codelines, behaviors_incall_codelines, hashmap_codelines


def get_behavior_call_codelines(
    graph: "HEBGraph",
    behaviors_incall_codelines: Set["Behavior"],
    indent: int = 1,
    with_overhead=True,
) -> List[str]:
    """Generate the codelines of a GeneratedBehavior call function.

    Args:
        graph (HEBGraph): HEBGraph from which to generate the behavior of.
        behaviors_incall_codelines (Set[Behavior]): Set of behavior to unroll directly
            instead of refering to.
        indent (int, optional): Indentation level. Defaults to 1.
        with_overhead (bool, optional): If True, adds the call function definition.
            Defaults to True.

    Returns:
        List[str]: Codelines of the GeneratedBehavior call function.
    """
    call_codelines = []
    if with_overhead:
        call_codelines.append(indent_str(indent) + "def __call__(self, observation):")
        indent += 1
    roots = get_roots(graph)
    return call_codelines + get_node_call_codelines(
        graph,
        roots[0],
        indent,
        behaviors_incall_codelines=behaviors_incall_codelines,
    )


def get_node_call_codelines(
    graph: "HEBGraph",
    node: Node,
    indent: int,
    behaviors_incall_codelines: Set["Behavior"],
) -> List[str]:
    """Generate codelines for an HEBGraph node recursively using the succesors.

    Args:
        graph (HEBGraph): HEBGraph containing the node.
        node (Node): Node to generate the call of.
        indent (int): Indentation level.
        behaviors_incall_codelines (Set[Behavior]): Set of behavior to unroll directly
            instead of refering to.
    Raises:
        NotImplementedError: Node is not an Action, FeatureCondition or Behavior.

    Returns:
        List[str]: Codelines of the node call.
    """
    node_codelines = []
    if isinstance(node, Action):
        action_name = node.name.split(BEHAVIOR_SEPARATOR)[-1]
        node_codelines.append(
            indent_str(indent) + f"return self.actions['{action_name}'](observation)"
        )
        return node_codelines
    if isinstance(node, FeatureCondition):
        var_name = f"edge_index_{indent-2}" if indent > 2 else "edge_index"
        fc_name = node.name.split(BEHAVIOR_SEPARATOR)[-1]
        node_codelines.append(
            indent_str(indent)
            + f"{var_name} = self.feature_conditions['{fc_name}'](observation)"
        )
        for i in [0, 1]:
            node_codelines.append(indent_str(indent) + f"if {var_name} == {i}:")
            successors = get_successors_with_index(graph, node, i)
            for succ_node in successors:
                node_codelines += get_node_call_codelines(
                    graph,
                    node=succ_node,
                    indent=indent + 1,
                    behaviors_incall_codelines=behaviors_incall_codelines,
                )
        return node_codelines
    if isinstance(node, Behavior):
        if node in behaviors_incall_codelines:
            if node.name in graph.all_behaviors:
                node = graph.all_behaviors[node.name]
            return get_behavior_call_codelines(
                node.graph,
                behaviors_incall_codelines=behaviors_incall_codelines,
                indent=indent,
                with_overhead=False,
            )
        default_line = f"return self.known_behaviors['{node.name}'](observation)"
        return [indent_str(indent) + default_line]
    raise NotImplementedError


def indent_str(indent_level: int, indent_amount: int = 4) -> str:
    """Gives a string indentation from a given indent level.

    Args:
        indent_level (int): Level of indentation.
        indent_amount (int, optional): Number of spaces per indent. Defaults to 4.

    Returns:
        str: Indentation string.
    """
    return " " * indent_level * indent_amount


def _to_camel_case(text: str) -> str:
    s = (
        text.replace("-", " ")
        .replace("_", " ")
        .replace("?", "")
        .replace("[", "")
        .replace("]", "")
        .replace("(", "")
        .replace(")", "")
        .replace(",", "")
    )
    s = s.split()
    if len(text) == 0:
        return text
    return s[0] + "".join(i.capitalize() for i in s[1:])


def _to_snake_case(text: str) -> str:
    text = text.replace("-", " ").replace("?", "")
    return "_".join(
        sub("([A-Z][a-z]+)", r" \1", sub("([A-Z]+)", r" \1", text)).split()
    ).lower()



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Metaheuristics used for building layouts """

import numpy as np


def simulated_annealing(
    initial,
    energy,
    neighbor,
    max_iterations: int = 1000,
    initial_temperature: float = 5,
    max_iters_without_new: int = np.inf,
    verbose: bool = 1,
):

    """Perform simulated annealing metaheuristic on an energy using a neighboring.

    See https://en.wikipedia.org/wiki/Simulated_annealing for more details.

    Args:
        initial: Initial variable position.
        energy: Function giving the energy (aka cost) of a given variable position.
        neighbor: Function giving a neighbor of a given variable position.
        max_iterations: Maximum number of iterations.
        initial_temperature: Initial temperature parameter, more is more random search.
        max_iters_without_new: Maximum number of iterations without a new best position.

    """

    def prob_keep(temperature, delta_e):
        return min(1, np.exp(delta_e / temperature))

    state = initial
    energy_pos = energy(state)
    iters_without_new = 0
    for k in range(max_iterations):
        new_state = neighbor(state)
        new_energy = energy(new_state)
        temperature = initial_temperature / (k + 1)
        iters_without_new += 1
        prob = prob_keep(temperature, energy_pos - new_energy)
        if np.random.random() < prob:
            if verbose == 1:
                print(
                    f"{k}\t({prob:.0%})\t{energy_pos:.2f}->{new_energy:.2f}", end="\r"
                )
            state, energy_pos = new_state, new_energy
            iters_without_new = 0

        if iters_without_new >= max_iters_without_new:
            break
    if verbose == 1:
        print()
    return state



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Unit tests for the hebg.behavior module. """

import pytest
import pytest_check as check
from pytest_mock import MockerFixture

from hebg.behavior import Behavior


class TestBehavior:

    """Behavior"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize variables."""
        self.node = Behavior("behavior_name")

    def test_node_type(self):
        """should have 'behavior' as node_type."""
        check.equal(self.node.type, "behavior")

    def test_node_call(self, mocker: MockerFixture):
        """should use graph on call."""
        mocker.patch("hebg.behavior.Behavior.graph")
        self.node(None)
        check.is_true(self.node.graph.called)

    def test_build_graph(self):
        """should raise NotImplementedError when build_graph is called."""
        with pytest.raises(NotImplementedError):
            self.node.build_graph()

    def test_graph(self, mocker: MockerFixture):
        """should build graph and compute its levels if, and only if,
        the graph is not yet built.
        """
        mocker.patch("hebg.behavior.Behavior.build_graph")
        mocker.patch("hebg.behavior.compute_levels")
        self.node.graph
        check.is_true(self.node.build_graph.called)
        check.is_true(self.node.build_graph.called)

        mocker.patch("hebg.behavior.Behavior.build_graph")
        mocker.patch("hebg.behavior.compute_levels")
        self.node.graph
        check.is_false(self.node.build_graph.called)
        check.is_false(self.node.build_graph.called)



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Module for base Node classes. """

import dis
from typing import Any, List

import numpy as np


def bytecode_complexity(obj):
    """Compute the number of instructions in the bytecode of a given obj."""
    return len(list(dis.get_instructions(obj)))


class Node:

    NODE_TYPES = ("action", "feature_condition", "behavior", "empty")

    def __init__(
        self,
        name: str,
        node_type: str,
        complexity: int = None,
        image=None,
    ) -> None:
        """Base Node class for any HEBGraph.

        Args:
            name (str): A UNIQUE name representing this node.
            node_type (str): One of {action, feature_condition, behavior, empty}.
            complexity (int, optional): Given individual complexity of the node.
                If None, uses the number of bytecode instructions of init and call.
                Defaults to None.
            image (2d array, optional): Image to represent the node. Defaults to None.

        Raises:
            ValueError: If node_type has an unexpected value.
        """
        self.name = name
        self.image = image
        if node_type not in self.NODE_TYPES:
            raise ValueError(
                f"node_type ({node_type})"
                f"not in authorised node_types ({self.NODE_TYPES})."
            )
        self.type = node_type
        if complexity is not None:
            self.complexity = complexity
        else:
            self.complexity = bytecode_complexity(self.__init__)
            self.complexity += bytecode_complexity(self.__call__)

    def __call__(self, observation: Any) -> Any:
        raise NotImplementedError

    def __str__(self) -> str:
        return self.name

    def __eq__(self, o: object) -> bool:
        return self.name == str(o)

    def __hash__(self) -> int:
        return self.name.__hash__()

    def __repr__(self) -> str:
        return self.name


class Action(Node):

    """Node representing an action in an HEBGraph."""

    def __init__(self, action: Any, name: str = None, **kwargs) -> None:
        self.action = action
        super().__init__(self._get_name(name), "action", **kwargs)

    def _get_name(self, name):
        """Get the default name of the action if None is given."""
        return f"action {self.action}" if name is None else name

    def __call__(self, observation: Any) -> Any:
        return self.action


class StochasticAction(Action):

    """Node representing a stochastic choice between actions in an HEBGraph."""

    def __init__(
        self, actions: List[Action], probs: list, name: str, image=None
    ) -> None:
        super().__init__(actions, name=name, image=image)
        self.probs = probs

    def __call__(self, observation):
        selected_action = np.random.choice(self.action, p=self.probs)
        return selected_action(observation)


class FeatureCondition(Node):

    """Node representing a feature condition in an HEBGraph."""

    def __init__(self, name: str = None, **kwargs) -> None:
        super().__init__(name, "feature_condition", **kwargs)

    def __call__(self, observation: Any) -> int:
        raise NotImplementedError


class EmptyNode(Node):

    """Node representing an empty node in an HEBGraph."""

    def __init__(self, name: str) -> None:
        super().__init__(name, "empty")

    def __call__(self, observation: Any) -> int:
        return int(True)



import math
from typing import TYPE_CHECKING, Dict, Optional, Tuple

import numpy as np
from scipy.spatial import ConvexHull  # pylint: disable=no-name-in-module

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.legend_handler import HandlerPatch
from matplotlib.legend import Legend
from matplotlib.axes import Axes

from networkx import draw_networkx_edges

from hebg.unrolling import group_behaviors_points
from hebg.layouts import staircase_layout
from hebg.graph import draw_networkx_nodes_images

if TYPE_CHECKING:
    from hebg.node import Node
    from hebg.heb_graph import HEBGraph


def draw_hebgraph(
    graph: "HEBGraph",
    ax: Axes,
    pos: Optional[Dict["Node", Tuple[float, float]]] = None,
    fontcolor: str = "black",
    draw_hulls: bool = False,
    show_all_hulls: bool = False,
) -> Tuple["Axes", Dict["Node", Tuple[float, float]]]:
    if len(list(graph.nodes())) > 0:
        if pos is None:
            pos = staircase_layout(graph)
        draw_networkx_nodes_images(graph, pos, ax=ax, img_zoom=0.5)

        draw_networkx_edges(
            graph,
            pos,
            ax=ax,
            arrowsize=20,
            arrowstyle="-|>",
            min_source_margin=0,
            min_target_margin=10,
            node_shape="s",
            node_size=1500,
            edge_color=[color for _, _, color in graph.edges(data="color")],
        )

        legend = draw_graph_legend(graph, ax)
        plt.setp(legend.get_texts(), color=fontcolor)

        if draw_hulls:
            group_and_draw_hulls(graph, pos, ax, show_all_hulls=show_all_hulls)


def draw_graph_legend(graph: "HEBGraph", ax: Axes) -> Legend:
    used_node_types = [node_type for _, node_type in graph.nodes(data="type")]
    legend_patches = [
        mpatches.Patch(facecolor="none", edgecolor=color, label=node_type.capitalize())
        for node_type, color in graph.NODES_COLORS.items()
        if node_type in used_node_types and node_type in graph.NODES_COLORS
    ]
    used_edge_indexes = [index for _, _, index in graph.edges(data="index")]
    legend_arrows = [
        mpatches.FancyArrow(
            *(0, 0, 1, 0),
            facecolor=color,
            edgecolor="none",
            label=str(index) if index > 1 else f"{str(bool(index))} ({index})",
        )
        for index, color in graph.EDGES_COLORS.items()
        if index in used_edge_indexes and index in graph.EDGES_COLORS
    ]

    # Draw the legend
    legend = ax.legend(
        fancybox=True,
        framealpha=0,
        fontsize="x-large",
        loc="upper right",
        handles=legend_patches + legend_arrows,
        handler_map={
            # Patch arrows with fancy arrows in legend
            mpatches.FancyArrow: HandlerPatch(
                patch_func=lambda width, height, **kwargs: mpatches.FancyArrow(
                    *(0, 0.5 * height, width, 0),
                    width=0.2 * height,
                    length_includes_head=True,
                    head_width=height,
                    overhang=0.5,
                )
            ),
        },
    )

    return legend


def group_and_draw_hulls(graph: "HEBGraph", pos, ax: Axes, show_all_hulls: bool):
    grouped_points = group_behaviors_points(pos, graph)
    if not show_all_hulls:
        key_count = {key[-1]: 0 for key in grouped_points}
        for key in grouped_points:
            key_count[key[-1]] += 1
        grouped_points = {
            key: points
            for key, points in grouped_points.items()
            if key_count[key[-1]] > 1 and (len(key) == 1 or key[-1] != key[-2])
        }

    for group_key, points in grouped_points.items():
        stretch = 0.5 - 0.05 * (len(group_key) - 1)
        if len(points) >= 3:
            draw_convex_hull(points, ax, stretch=stretch, lw=3, color="orange")


def draw_convex_hull(points, ax: "Axes", stretch=0.3, n_points=30, **kwargs):
    points = np.array(points)
    convh = ConvexHull(points)  # Get the first convexHull (speeds up the next process)
    points = buffer_points(points[convh.vertices], stretch=stretch, samples=n_points)

    hull = ConvexHull(points)
    hull_cycle = np.concatenate((hull.vertices, hull.vertices[:1]))
    ax.plot(points[hull_cycle, 0], points[hull_cycle, 1], **kwargs)


def buffer_points(inside_points, stretch, samples):
    new_points = []
    for point in inside_points:
        new_points += points_in_circum(point, stretch, samples)
    new_points = np.array(new_points)
    hull = ConvexHull(new_points)
    return new_points[hull.vertices]


def points_in_circum(points, radius, samples=100):
    return [
        (
            points[0] + math.cos(2 * math.pi / samples * x) * radius,
            points[1] + math.sin(2 * math.pi / samples * x) * radius,
        )
        for x in range(0, samples + 1)
    ]



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Behavior of HEBGraphs with empty nodes. """

import pytest_check as check

from hebg.node import Action
from tests.examples.behaviors import (
    E_F_A_Behavior,
    E_A_Behavior,
    F_E_A_Behavior,
    E_E_A_Behavior,
)
from tests.examples.feature_conditions import ThresholdFeatureCondition


def test_e_a_graph():
    """(E-A) Empty nodes should skip to successor."""
    action_id = 42
    behavior = E_A_Behavior("E_A", Action(action_id))
    check.equal(behavior(None), action_id)


def test_e_f_a_graph():
    """(E-F-A) Empty should orient path properly in chain with Feature condition."""
    behavior = E_F_A_Behavior("E_F_A")
    check.equal(behavior(-1), 0)
    check.equal(behavior(1), 1)


def test_f_e_a_graph():
    """(F-E-A) Feature condition should orient path properly in chain with Empty."""
    behavior = F_E_A_Behavior("F_E_A")
    check.equal(behavior(1), 0)
    check.equal(behavior(-1), 1)


def test_e_e_a_graph():
    """(E-E-A) Empty should orient path properly in double chain."""
    behavior = E_E_A_Behavior("E_E_A")
    check.equal(behavior(None), 0)



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Utility functions for complexity computation. """

from copy import deepcopy


def update_sum_dict(dict1: dict, dict2: dict):
    """Give the sum of two dictionaries."""
    dict1, dict2 = deepcopy(dict1), deepcopy(dict2)
    for key, val in dict2.items():
        if not isinstance(val, dict):
            try:
                dict1[key] += val
            except KeyError:
                dict1[key] = val
    return dict1



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>

""" Module for utility computation methods. """

from hebg.metrics.utility.binary_utility import binary_graphbased_utility



# HEBGraph for explainable hierarchical reinforcement learning
# Copyright (C) 2021-2022 Mathïs FEDERICO <https://www.gnu.org/licenses/>
# pylint: disable=protected-access

""" Additional utility functions for networkx graphs. """

from typing import TYPE_CHECKING, Any, Dict

from matplotlib.axes import Axes
from matplotlib.offsetbox import AnnotationBbox, OffsetImage
from networkx import DiGraph

if TYPE_CHECKING:
    from hebg.heb_graph import HEBGraph
    from hebg.node import Node


def get_roots(graph: DiGraph):
    """Finds roots in a DiGraph.

    Args:
        graph: A networkx DiGraph.

    Returns:
        List of root nodes.

    """
    roots = []
    for node in graph.nodes():
        if len(list(graph.predecessors(node))) == 0:
            roots.append(node)
    return roots


def get_nodes_by_level(graph: DiGraph) -> Dict[int, Any]:
    """Get the dictionary of nodes by level.

    Requires nodes to have a 'level' attribute.

    Args:
        graph: A networkx DiGraph.

    Returns:
        Dictionary of nodes by level.

    """
    nodes_by_level = {}
    for node in graph.nodes():
        level = graph.nodes[node]["level"]
        try:
            nodes_by_level[level].append(node)
        except KeyError:
            nodes_by_level[level] = [node]

    graph.graph["nodes_by_level"] = nodes_by_level
    graph.graph["depth"] = max(level for level in nodes_by_level)
    return nodes_by_level


def compute_levels(graph: DiGraph):
    """Compute the hierachical levels of all DiGraph nodes.

    Adds the attribute 'level' to each node in the given graph.
    Adds the attribute 'nodes_by_level' to the given graph.
    Adds the attribute 'depth' to the given graph.

    Args:
        graph: A networkx DiGraph.

    Returns:
        Dictionary of nodes by level.

    """

    def _compute_level_dependencies(graph: DiGraph, node):
        predecessors = list(graph.predecessors(node))
        if len(predecessors) == 0:
            graph.nodes[node]["level"] = 0
            return True

        pred_level_by_index = {}
        for pred in predecessors:
            index = graph.edges[pred, node]["index"]
            try:
                pred_level = graph.nodes[pred]["level"]
            except KeyError:
                pred_level = None

            if index in pred_level_by_index:
                pred_level_by_index[index].append(pred_level)
            else:
                pred_level_by_index[index] = [pred_level]

        min_level_by_index = []
        for index, level_list in pred_level_by_index.items():
            level_list_wo_none = [l for l in level_list if l is not None]
            if len(level_list_wo_none) == 0:
                return False
            min_level_by_index.append(min(level_list_wo_none))
        level = 1 + max(min_level_by_index)
        graph.nodes[node]["level"] = level
        return True

    for _ in range(len(graph.nodes())):
        all_nodes_have_level = True
        incomplete_nodes = []
        for node in graph.nodes():
            incomplete = not _compute_level_dependencies(graph, node)
            if incomplete:
                incomplete_nodes.append(node)
                all_nodes_have_level = False
        if all_nodes_have_level:
            break

    if not all_nodes_have_level:
        raise ValueError(
            "Could not attribute levels to all nodes. "
            f"Incomplete nodes: {incomplete_nodes}"
        )

    return get_nodes_by_level(graph)


def compute_edges_color(graph: DiGraph):
    """Compute the edges colors of a leveled graph for readability.

    Requires nodes to have a 'level' attribute.
    Adds the attribute 'color' and 'linestyle' to each edge in the given graph.
    Nodes with a lot of successors will have more transparent edges.
    Edges going from high to low level will be dashed.

    Args:
        graph: A networkx DiGraph.

    """
    alphas = [1, 1, 0.9, 0.8, 0.7, 0.5, 0.4, 0.3]
    for node in graph.nodes():
        successors = list(graph.successors(node))
        for succ in successors:
            alpha = 0.2
            if graph.nodes[node]["level"] < graph.nodes[succ]["level"]:
                if len(successors) < len(alphas):
                    alpha = alphas[len(successors) - 1]
            else:
                graph.edges[node, succ]["linestyle"] = "dashed"
            if isinstance(graph.edges[node, succ]["color"], list):
                graph.edges[node, succ]["color"][3] = alpha


def draw_networkx_nodes_images(
    graph: DiGraph, pos, ax: Axes, img_zoom: float = 1, **kwargs
):
    """Draw nodes images of a networkx DiGraph on a given matplotlib ax.

    Requires nodes to have attributes 'image' for the node image and 'color' for the border color.

    Args:
        graph: A networkx DiGraph.
        pos: Layout positions of the graph.
        ax: A matplotlib Axes.
        img_zoom (Optional): Zoom to apply to images.

    """
    for n in graph:
        img = graph.nodes[n]["image"]
        color = graph.nodes[n]["color"]
        if img is not None:
            min_dim = min(img.shape[:2])
            min_ax_shape = min(ax._position.width, ax._position.height)
            zoom = 100 * img_zoom * min_ax_shape / min_dim
            imagebox = OffsetImage(img, zoom=zoom)
            imagebox = AnnotationBbox(
                imagebox, pos[n], frameon=True, box_alignment=(0.5, 0.5)
            )

            imagebox.patch.set_facecolor("None")
            imagebox.patch.set_edgecolor(color)
            imagebox.patch.set_linewidth(3)
            imagebox.patch.set_boxstyle("round", pad=0.15)
            ax.add_artist(imagebox)
        else:
            # If no image is found, draw label instead
            (x, y) = pos[n]
            label = str(n)
            ax.text(
                x,
                y,
                label,
                size=kwargs.get("font_size", 12),
                color=kwargs.get("font_color", "k"),
                family=kwargs.get("font_family", "sans-serif"),
                weight=kwargs.get("font_weight", "normal"),
                alpha=kwargs.get("alpha"),
                horizontalalignment=kwargs.get("horizontalalignment", "center"),
                verticalalignment=kwargs.get("verticalalignment", "center"),
                transform=ax.transData,
                bbox=kwargs.get("bbox"),
                clip_on=True,
            )


def get_successors_with_index(graph: "HEBGraph", node: "Node", next_edge_index: int):
    succs = graph.successors(node)
    next_nodes = []
    for next_node in succs:
        if int(graph.edges[node, next_node]["index"]) == next_edge_index:
            next_nodes.append(next_node)
    if len(next_nodes) == 0:
        raise ValueError(
            f"FeatureCondition {node} returned index {next_edge_index}"
            f" but {next_edge_index} was not found as an edge index"
        )
    return next_nodes



from typing import Optional, Tuple, Any

import pytest
import pytest_check as check

from hebg import Action, FeatureCondition, Behavior

from tests.examples.behaviors import (
    FundamentalBehavior,
    F_A_Behavior,
    F_F_A_Behavior,
    build_binary_sum_behavior,
)
from tests.examples.feature_conditions import ThresholdFeatureCondition


class TestABehavior:
    """(A) Fundamental behaviors (single Action node) should return action call."""

    @pytest.fixture(autouse=True)
    def setup(self):
        self.behavior = FundamentalBehavior(Action(42))

    def test_source_codegen(self):
        source_code = self.behavior.graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class Action42Behavior(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        return self.actions['action 42'](observation)",
            )
        )
        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )

    def test_exec_codegen(self):
        check_execution_for_values(self.behavior, "Action42Behavior", (1, -1))


class TestFABehavior:
    """(F-A) Feature condition should generate if/else condition."""

    @pytest.fixture(autouse=True)
    def setup(self):
        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)
        actions = {0: Action(0), 1: Action(1)}
        self.behavior = F_A_Behavior("Is above_zero", feature_condition, actions)

    def test_source_codegen(self):
        source_code = self.behavior.graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class IsAboveZero(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['Greater or equal to 0 ?'](observation)",
                "        if edge_index == 0:",
                "            return self.actions['action 0'](observation)",
                "        if edge_index == 1:",
                "            return self.actions['action 1'](observation)",
            )
        )

        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )

    def test_exec_codegen(self):
        check_execution_for_values(self.behavior, "IsAboveZero", (1, -1))


class TestFFABehavior:
    """(F-F-A) Chained FeatureConditions should condition should generate nested if/else."""

    @pytest.fixture(autouse=True)
    def setup(self):
        self.behavior = F_F_A_Behavior("scalar classification ]-1,0,1[ ?")

    def test_source_codegen(self):
        source_code = self.behavior.graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class ScalarClassification101(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['Greater or equal to 0 ?'](observation)",
                "        if edge_index == 0:",
                "            edge_index_1 = self.feature_conditions['Greater or equal to -1 ?'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 0'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.actions['action 1'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['Lesser or equal to 1 ?'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 3'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.actions['action 2'](observation)",
            )
        )

        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )

    def test_exec_codegen(self):
        check_execution_for_values(
            self.behavior, "ScalarClassification101", (2, 1, -1, -2)
        )


class TestFBBehavior:
    """(F-BA) Behaviors should be unrolled by default if they appear only once and have a graph."""

    @pytest.fixture(autouse=True)
    def setup(self):
        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)
        actions = {0: Action(0), 1: Action(1)}
        sub_behavior = F_A_Behavior("Is above_zero", feature_condition, actions)

        feature_condition = ThresholdFeatureCondition(relation="<=", threshold=1)
        actions = {0: Action(0), 1: sub_behavior}
        self.behavior = F_A_Behavior("Is between 0 and 1 ?", feature_condition, actions)

    def test_source_codegen(self):
        source_code = self.behavior.graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class IsBetween0And1(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['Lesser or equal to 1 ?'](observation)",
                "        if edge_index == 0:",
                "            return self.actions['action 0'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['Greater or equal to 0 ?'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 0'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.actions['action 1'](observation)",
            )
        )

        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )

    def test_exec_codegen(self):
        check_execution_for_values(self.behavior, "IsBetween0And1", (-1, 0, 1, 2))


class TestFBBehaviorNameRef:
    """(F-BA) Behaviors should work with only name reference to behavior,
    but will expect behavior to be given, even when unrolled."""

    @pytest.fixture(autouse=True)
    def setup(self):
        feature_condition = ThresholdFeatureCondition(relation="<=", threshold=1)
        actions = {0: Action(0), 1: Behavior("Is above_zero")}
        self.behavior = F_A_Behavior("Is between 0 and 1 ?", feature_condition, actions)

    @pytest.mark.filterwarnings("ignore:Could not load graph for behavior")
    def test_source_codegen_by_ref(self):
        unrolled_source_code = self.behavior.graph.generate_source_code()
        source_code = self.behavior.graph.unrolled_graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "# Require 'Is above_zero' behavior to be given.",
                "class IsBetween0And1(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['Lesser or equal to 1 ?'](observation)",
                "        if edge_index == 0:",
                "            return self.actions['action 0'](observation)",
                "        if edge_index == 1:",
                "            return self.known_behaviors['Is above_zero'](observation)",
            )
        )

        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )
        check.equal(
            unrolled_source_code,
            expected_source_code,
            msg=_unidiff_output(unrolled_source_code, expected_source_code),
        )

    def test_source_codegen_in_all_behavior(self):
        """When the behavior is found in 'all_behaviors'
        it should used the found behavior for codegen."""
        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)
        actions = {0: Action(0), 1: Action(1)}
        sub_behavior = F_A_Behavior("Is above_zero", feature_condition, actions)
        self.behavior.graph.all_behaviors["Is above_zero"] = sub_behavior
        source_code = self.behavior.graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class IsBetween0And1(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['Lesser or equal to 1 ?'](observation)",
                "        if edge_index == 0:",
                "            return self.actions['action 0'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['Greater or equal to 0 ?'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 0'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.actions['action 1'](observation)",
            )
        )

        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )

    def test_exec_codegen(self):
        """When the behavior is found in 'all_behaviors'
        it should used the found behavior for graph call."""
        feature_condition = ThresholdFeatureCondition(relation=">=", threshold=0)
        actions = {0: Action(0), 1: Action(1)}
        sub_behavior = F_A_Behavior("Is above_zero", feature_condition, actions)
        self.behavior.graph.all_behaviors["Is above_zero"] = sub_behavior
        check_execution_for_values(
            self.behavior,
            "IsBetween0And1",
            (-1, 0, 1, 2),
            known_behaviors={"Is above_zero": sub_behavior},
        )


class TestNestedBehaviorReuse:
    """Behaviors should be rolled when they are used mutliple times in nested subgraphs."""

    @pytest.fixture(autouse=True)
    def setup(self):
        feature_condition = FeatureCondition(name="fc1")
        actions = {0: Action(0), 1: Action(1)}
        behavior_0 = F_A_Behavior("behavior 0", feature_condition, actions)

        feature_condition = FeatureCondition(name="fc2")
        actions = {0: Action(0), 1: behavior_0}
        behavior_1 = F_A_Behavior("behavior 1", feature_condition, actions)

        feature_condition = FeatureCondition(name="fc3")
        actions = {0: behavior_0, 1: behavior_1}
        self.behavior = F_A_Behavior("behavior 2", feature_condition, actions)

    def test_nested_reuse_codegen(self):
        source_code = self.behavior.graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class Behavior0(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['fc1'](observation)",
                "        if edge_index == 0:",
                "            return self.actions['action 0'](observation)",
                "        if edge_index == 1:",
                "            return self.actions['action 1'](observation)",
                "class Behavior2(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['fc3'](observation)",
                "        if edge_index == 0:",
                "            return self.known_behaviors['behavior 0'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['fc2'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 0'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.known_behaviors['behavior 0'](observation)",
                "BEHAVIOR_TO_NAME = {",
                "    'behavior 0': Behavior0,",
                "}",
            )
        )

        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )


class TestFundamentalBehaviorReuse:
    """Fundamental Behaviors should never be abstracted."""

    @pytest.fixture(autouse=True)
    def setup(self):
        behavior_0 = FundamentalBehavior(Action(1))

        feature_condition = FeatureCondition(name="fc2")
        actions = {0: Action(0), 1: behavior_0}
        behavior_1 = F_A_Behavior("behavior 1", feature_condition, actions)

        feature_condition = FeatureCondition(name="fc3")
        actions = {0: behavior_0, 1: behavior_1}
        self.behavior = F_A_Behavior("behavior 2", feature_condition, actions)

    def test_nested_reuse_codegen(self):
        source_code = self.behavior.graph.generate_source_code()
        expected_source_code = "\n".join(
            (
                "from hebg.codegen import GeneratedBehavior",
                "",
                "class Behavior2(GeneratedBehavior):",
                "    def __call__(self, observation):",
                "        edge_index = self.feature_conditions['fc3'](observation)",
                "        if edge_index == 0:",
                "            return self.actions['action 1'](observation)",
                "        if edge_index == 1:",
                "            edge_index_1 = self.feature_conditions['fc2'](observation)",
                "            if edge_index_1 == 0:",
                "                return self.actions['action 0'](observation)",
                "            if edge_index_1 == 1:",
                "                return self.actions['action 1'](observation)",
            )
        )

        check.equal(
            source_code,
            expected_source_code,
            msg=_unidiff_output(source_code, expected_source_code),
        )


class TestFBBBehavior:
    """(F-B-B) Behaviors should only be added once as a class."""

    @pytest.fixture(autouse=True)
    def setup(self):
        self.behavior = build_binary_sum_behavior()

    def test_classes_in_codegen(self):
        source_code = self.behavior.graph.generate_source_code()
        expected_classes = [
            "IsX1InBinary",  # Only this one is used twice
            "IsSumOfLast3Binary2",
        ]

        for expected_class in expected_classes:
            check.equal(
                source_code.count(f"class {expected_class}"),
                1,
                msg=f"Missing or duplicated class: {expected_class}\n{source_code}",
            )

    def test_exec_codegen(self):
        check_execution_for_values(
            self.behavior, "IsSumOfLast3Binary2", (0, 1, 3, 5, 15)
        )


def check_execution_for_values(
    behavior: Behavior,
    class_name: str,
    values: Tuple[Any],
    known_behaviors: Optional[dict] = None,
):
    generated_source_code = behavior.graph.generate_source_code()
    exec(generated_source_code)
    CodeGenPolicy = locals()[class_name]

    actions, feature_conditions, behaviors = separate_nodes_by_type(behavior)

    _behaviors = behaviors.copy()
    while len(_behaviors) > 0:
        _, sub_behavior = _behaviors.popitem()
        if sub_behavior in behavior.graph.all_behaviors:
            sub_behavior = behavior.graph.all_behaviors[sub_behavior]
        sub_actions, sub_feature_conditions, sub_behaviors = separate_nodes_by_type(
            sub_behavior
        )
        actions.update(sub_actions)
        feature_conditions.update(sub_feature_conditions)
        _behaviors.update(sub_behaviors)
        behaviors.update(sub_behaviors)

    known_behaviors = known_behaviors if known_behaviors is not None else {}
    behaviors.update(known_behaviors)

    behavior_rebuilt = CodeGenPolicy(
        actions=actions,
        feature_conditions=feature_conditions,
        behaviors=behaviors,
    )

    for val in values:
        check.equal(behavior(val), behavior_rebuilt(val))


def separate_nodes_by_type(behavior: Behavior):
    actions = {
        node.name: node for node in behavior.graph.nodes if isinstance(node, Action)
    }
    feature_conditions = {
        node.name: node
        for node in behavior.graph.nodes
        if isinstance(node, FeatureCondition)
    }
    behaviors = {
        node.name: node for node in behavior.graph.nodes if isinstance(node, Behavior)
    }
    return actions, feature_conditions, behaviors


def _unidiff_output(expected: str, actual: str):
    """
    Helper function. Returns a string containing the unified diff of two multiline strings.
    """

    import difflib

    expected = [line for line in expected.split("\n")]
    actual = [line for line in actual.split("\n")]

    diff = difflib.unified_diff(expected, actual)

    return "\n".join(diff)
